"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RoutingConfigurer = void 0;
const tslib_1 = require("tslib");
//    Copyright 2018 Quirino Brizi [quirino.brizi@gmail.com]
//
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
//
//        http://www.apache.org/licenses/LICENSE-2.0
//
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.
;
('use strict');
const apicache_plus_1 = tslib_1.__importDefault(require("apicache-plus"));
const dino_core_1 = require("dino-core");
const api_validator_1 = require("./api.validator");
const helper_1 = require("./helper");
const application_state_monitor_1 = require("./monitoring/application.state.monitor");
const cache_performance_monitor_1 = require("./monitoring/cache.performance.monitor");
const co2_emission_monitor_1 = require("./monitoring/co2.emission.monitor");
const route_handler_1 = require("./route.handler");
const error_handler_1 = require("./error.handler");
const filesystem_descriptor_provider_1 = require("./openapi/filesystem.descriptor.provider");
const after_handler_middleware_1 = require("./middlewares/after.handler.middleware");
const before_handler_middleware_1 = require("./middlewares/before.handler.middleware");
const ProjectionsMiddleware_1 = require("./middlewares/impl/after/ProjectionsMiddleware");
const SendResponseMiddleware_1 = require("./middlewares/impl/after/SendResponseMiddleware");
const SecurityMiddleware_1 = require("./middlewares/impl/before/SecurityMiddleware");
const RuntimeContext_1 = require("./RuntimeContext");
const DESCRIPTOR_PROVIDER_MAP = Object.freeze({ filesystem: filesystem_descriptor_provider_1.FilesystemDescriptorProvider });
/**
 * Configure express routing based on swagger configuration.
 *
 * @typedef RoutingConfigurer
 * @public
 */
class RoutingConfigurer extends dino_core_1.Component {
    server;
    applicationStateMonitor;
    environment;
    applicationContext;
    runtimeContext;
    constructor({ applicationContext, environment }) {
        super();
        this.server = undefined;
        this.environment = environment;
        this.applicationContext = applicationContext;
        this.runtimeContext = new RuntimeContext_1.RuntimeContext(environment);
        applicationContext.register('dinoRuntimeContext', dino_core_1.ComponentDescriptor.createFromValue('dinoRuntimeContext', this.runtimeContext, dino_core_1.Scope.SINGLETON));
    }
    async postConstruct() {
        const express = this.applicationContext.require('express');
        const app = express();
        const swaggerUi = this.applicationContext.require('swagger-ui-express');
        const swaggerParser = this.applicationContext.require('swagger-parser');
        app.use(express.json());
        app.disable('x-powered-by');
        this.extend(app);
        const router = express.Router();
        const apis = await swaggerParser.dereference(this.resolveDescriptorProvider().provide());
        const apiValidator = new api_validator_1.ApiValidator(this.environment);
        apiValidator.init(apis);
        const errorHandler = this.errorHandler();
        const docsPath = this.getVariable('dino:server:docs', '/api-docs');
        router.use(docsPath, swaggerUi.serve);
        router.use(docsPath, swaggerUi.setup(apis, { explorer: false }));
        this.setupMonitoringEndpointsIfRequired(router);
        const monitoringMiddlewares = this.defineTopLevelMonitorsIfRequired();
        Object.keys(apis.paths).forEach((path) => {
            const currentPath = apis.paths[path];
            Object.keys(currentPath).forEach((method) => {
                const apiPath = helper_1.Helper.normalizePath(path);
                const api = currentPath[method];
                const responseValidator = apiValidator.getResponseValidator(method, path);
                const requestHandler = this.requestHandler(api, responseValidator);
                // defines middlewares proposing built-ins and concatenating lesser ones
                // as per express style the user will be able to interrupt the chain just
                // sending a response, this at current stage is not a concern but a better
                // middleware structure that avoid this kind if hacks should be put in place.
                // this is tracked on https://gitlab.com/codesketch/dino-express/issues/7
                const userDefinedMiddlewares = this.middlewares(api, apis.components) || [];
                const beforeMiddlewares = [apiValidator.getRequestValidator(method, path)]
                    .concat(userDefinedMiddlewares.before)
                    .concat(monitoringMiddlewares.before);
                const afterMiddlewares = []
                    .concat(userDefinedMiddlewares.after)
                    .concat(monitoringMiddlewares.after)
                    .concat([new SendResponseMiddleware_1.SendResponseMiddleware()]);
                const middlewares = []
                    .concat(beforeMiddlewares)
                    .concat([requestHandler, errorHandler])
                    .concat(afterMiddlewares)
                    .map((middleware) => {
                    if (helper_1.Helper.instanceOf(middleware, after_handler_middleware_1.AfterHandlerMiddleware) || helper_1.Helper.instanceOf(middleware, before_handler_middleware_1.BeforeHandlerMiddleware)) {
                        return middleware.handle.bind(this);
                    }
                    return middleware;
                });
                router[method](apiPath, middlewares);
            });
        });
        this.startServer(app, router);
    }
    startServer(app, router) {
        app.use(this.getVariable('dino:server:path', '/'), router);
        const serverType = this.environment.getOrDefault('dino:server:type', 'express');
        if (serverType === 'express') {
            let port = this.getVariable('dino:server:port', 3030);
            this.server = app.listen(port, () => {
                dino_core_1.Logger.info(`application started, listening on port ${port}`);
            });
        }
        else {
            // if serverless register the router so that it can be used at later stages
            this.applicationContext.register('dinoApiRouter', dino_core_1.ComponentDescriptor.createFromValue('router', router, dino_core_1.Scope.SINGLETON));
        }
    }
    resolveDescriptorProvider() {
        let answer;
        try {
            dino_core_1.Logger.info('try locate user defined descriptor provider from the application context');
            answer = this.applicationContext.resolve('descriptorProvider');
        }
        catch (e) {
            dino_core_1.Logger.info('user defined descriptor provider not found using default file system based one');
            const DescriptorProvider = DESCRIPTOR_PROVIDER_MAP[this.getVariable('dino:descriptor:provider', 'filesystem')];
            answer = new DescriptorProvider(this.environment);
        }
        return answer;
    }
    setupMonitoringEndpointsIfRequired(router) {
        let monitoringEnabled = this.runtimeContext.isMonitoringEnabled();
        if (monitoringEnabled) {
            this.applicationStateMonitor = new application_state_monitor_1.ApplicationStateMonitor(this.applicationContext);
            router.use('/monitor', this.applicationStateMonitor.middleware());
        }
    }
    /**
     * Allows to define top level monitors. The returned objects contains two arrays
     * before and after middlewares, the before middlewares are added at the very start
     * of the middleware chain while the after are added at the very end of the middleware
     * chain, both in the order that are defined.
     *
     * @returns {Object}
     * ```{
     *    before: [],
     *    after: []
     *  }```
     */
    defineTopLevelMonitorsIfRequired() {
        const answer = { before: [], after: [] };
        let co2MonitoringEnabled = this.runtimeContext.isCo2MonitoringEnabled();
        if (co2MonitoringEnabled) {
            if (co2MonitoringEnabled) {
                const co2Monitor = new co2_emission_monitor_1.CO2EmissionMonitor();
                this.addMonitor(co2Monitor);
                answer.before.push(co2Monitor.requestMiddleware.bind(co2Monitor));
                answer.after.push(co2Monitor.responseMiddleware.bind(co2Monitor));
            }
        }
        return answer;
    }
    /**
     * Allows to add a monitor
     * @param {Monitor} monitor
     */
    addMonitor(monitor) {
        if (dino_core_1.ObjectHelper.isDefined(this.applicationStateMonitor)) {
            this.applicationStateMonitor.addMonitor(monitor);
        }
        return this;
    }
    preDestroy() {
        if (dino_core_1.ObjectHelper.isDefined(this.server)) {
            this.server.close();
        }
    }
    /**
     * Hook method allow to customize the express instance. by default it add the json middleware only.
     * This method can be overwritten in order to provide more tailored configuration of express instance.
     *
     * @param express the express instance
     * @protected
     */
    extend(_express) { }
    /**
     * Hook allows to provide an express error handler, by default a JSON error handler is used.
     * @protected
     */
    errorHandler() {
        return error_handler_1.ErrorHandler.instance(this.environment);
    }
    /**
     * Allows users to provide custom middlewares the middlewares will be inserted on the
     * chain as provided by the user but between the built-in API validation and route
     * handler and error handler.
     * @param {any} api the swagger API definition
     * @param {any} components the components OpenAPI definition
     * @returns a collection of middlewares split between before and after request handler
     *
     * @protected
     */
    middlewares(api, components) {
        return {
            before: [new SecurityMiddleware_1.SecurityMiddleware(this.environment), this.getCache(api, components)],
            after: [new ProjectionsMiddleware_1.ProjectionsMiddleware()]
        };
    }
    /**
     * Hook allow to define a custom request handler middleware. If no middleware is
     * returned the default will be used.
     * @param {any} api the OpenAPI definition
     * @param {function} responseValidator the validator for the response expected by the OpenAPI definition.
     * @returns {Function}
     *
     * @protected
     */
    requestHandler(api, responseValidator) {
        const routeHandler = new route_handler_1.RouteHandler(this.applicationContext, api, responseValidator, this.environment);
        return routeHandler.handle.bind(routeHandler);
    }
    /**
     * Retrieve the value associated with the key or the default.
     * @param {String} key the key for the value to retrieve
     * @param {any} _default the default value
     *
     * @private
     */
    getVariable(key, _default) {
        return helper_1.Helper.getVariable(this.environment, key, _default);
    }
    /**
     * Get the cache middleware configured as requested for the current route
     * @param {any} cache the cache configuration
     * @returns {Function} the cache middleware.
     *
     * @private
     */
    getCache(api, _components) {
        dino_core_1.Logger.debug(`configuring cache for API ${api.operationId}`);
        let answer = (_req, _res, next) => next();
        if (api.cache) {
            dino_core_1.Logger.debug(`configuring cache ttl to ${api.cache.ttl}`);
            const cacheInstance = apicache_plus_1.default.newInstance({
                debug: api.cache.debug || false,
                trackPerformance: api.cache.trackPerformance || true,
                isBypassable: api.cache.isBypassable || true
            });
            answer = cacheInstance.middleware(api.cache.ttl || '30 minutes', this.cacheOnlySuccessResponsesForGet);
            this.addMonitor(new cache_performance_monitor_1.CachePerformanceMonitor().setApiName(api.operationId).setCacheInstance(cacheInstance));
        }
        return answer;
    }
    /**
     * Define a function that discriminate what responses should be cached. In this case only responses
     * with 200 response code from GET requests.
     * @returns {Function} the cache toggle function
     *
     * @private
     */
    cacheOnlySuccessResponsesForGet(req, res) {
        dino_core_1.Logger.debug(`${res.statusCode} ${req.method}`);
        return res.statusCode == 200 && req.method.toLowerCase() == 'get';
    }
}
exports.RoutingConfigurer = RoutingConfigurer;
//# sourceMappingURL=routing.configurer.js.map